#include <stdint.h>
#include "nrf_sdh.h"
#include "nrf_sdh_ant.h"
#include "nrfx_rtc.h"
#include "app_error.h"
#include "ant_interface.h"
#include "ant_parameters.h"
#include "ant_channel_config.h"

// ANT channel config.
#define CHAN_ID_DEV_TYPE               1                            ///< Channel ID, Device Type. 
#define CHAN_ID_DEV_NUM                1                            ///< Channel ID, Device Number. 
#define CHAN_ID_TRANS_TYPE             1                            ///< Channel ID, Transmission type.
#define CHAN_PERIOD                    9687                         ///< Channel Period (in 32 kHz counts).
#define RF_FREQ                        66                           ///< RF Frequency: 2466 MHz. 
#define ANT_CHANNEL_DEFAULT_NETWORK    0x00                         ///< ANT Channel Network.
#define ANT_CHANNEL_NUMBER             0x00                         ///< ANT Channel Number.
#define EXT_TYPE                       0                            ///< Extended assign.
#define APP_ANT_OBSERVER_PRIO          1                            ///< Application's ANT observer priority. You shouldn't need to modify this value.
 
// Timesync Page Numbers
#define TIME_SYNC_PAGE                 0x01 
#define INVALIDATION_BYTE              0xFF   

// Misc. defines
#define FIRST_BYTE(a)                  ((a) & 0xFF)
#define SECOND_BYTE(a)                 ((a >> 8) & 0xFF)
#define THIRD_BYTE(a)                  ((a >> 16) & 0xFF)
#define MSGS_BETWEEN_TIMESYNCS         3

// RTC1 instance
const nrfx_rtc_t m_rtc = NRFX_RTC_INSTANCE(1);

// Message counter
uint8_t msg_counter = 0;

/** @brief: Function for handling the RTC1 interrupts.
 */
static void rtc_dummy_handler(nrfx_rtc_int_type_t int_type)
{
}

/** @brief Function for initialization and configuration of RTC driver instance.
 */
void rtc_init(void)
{
    uint32_t err_code;

    // Initialize RTC instance
    nrfx_rtc_config_t config = NRFX_RTC_DEFAULT_CONFIG;
    err_code = nrfx_rtc_init(&m_rtc, &config, rtc_dummy_handler);
    APP_ERROR_CHECK(err_code);

    // Power on RTC instance
    nrfx_rtc_enable(&m_rtc);
}

/**@brief Function for configuring the ANT channel.
 */
void antmaster_ant_setup(void)
{
    uint32_t err_code;
    ANT_TIME_SYNC_CONFIG syncConfig;

    // Configure ANT Time Synchronization using RTC1
    syncConfig.ucTimeBase           = ANT_TIME_BASE_ALT1; 
    syncConfig.bInvalidationEnabled = true; 
    syncConfig.ucInvalidationByte   = INVALIDATION_BYTE;
    err_code                        = sd_ant_time_sync_config_set(&syncConfig);
    APP_ERROR_CHECK(err_code);

    // ANT Channel init
    ant_channel_config_t channel_config =
    {
        .channel_number    = ANT_CHANNEL_NUMBER,
        .channel_type      = CHANNEL_TYPE_MASTER,
        .ext_assign        = EXT_TYPE,
        .rf_freq           = RF_FREQ,
        .transmission_type = CHAN_ID_TRANS_TYPE,
        .device_type       = CHAN_ID_DEV_TYPE,
        .device_number     = CHAN_ID_DEV_NUM,
        .channel_period    = CHAN_PERIOD,
        .network_number    = ANT_CHANNEL_DEFAULT_NETWORK,
    };

    err_code = ant_channel_init(&channel_config);
    APP_ERROR_CHECK(err_code);

    // Configure Adv. Burst
    uint8_t adv_burst_config[] = {ADV_BURST_MODE_ENABLE, ADV_BURST_MODES_SIZE_24_BYTES, 0, 0, 0, 0, 0, 0};
    err_code = sd_ant_adv_burst_config_set(adv_burst_config, sizeof(adv_burst_config));
    APP_ERROR_CHECK(err_code);

    // Channel open
    err_code = sd_ant_channel_open(ANT_CHANNEL_NUMBER);
    APP_ERROR_CHECK(err_code);

    // Send the first ACK msg
    uint8_t buffer[8] = {0};
    err_code = sd_ant_acknowledge_message_tx(ANT_CHANNEL_NUMBER,
                                             ANT_STANDARD_DATA_PAYLOAD_SIZE,
                                             buffer);
    APP_ERROR_CHECK(err_code);
}

/** @brief Function for setting payload for ANT timesync message and sending it.
 */
void send_timesync_message() 
{
    uint32_t err_code;
    uint8_t  message_payload[ANT_STANDARD_DATA_PAYLOAD_SIZE] = {0};
    uint32_t rtc_counter;

    // Retrieve current RTC counter
    rtc_counter = nrfx_rtc_counter_get(&m_rtc);

    // Time sync page
    message_payload[0] = TIME_SYNC_PAGE;

    // Place RTC current value in User Data.
    message_payload[3] = FIRST_BYTE(rtc_counter);
    message_payload[4] = SECOND_BYTE(rtc_counter);
    message_payload[5] = THIRD_BYTE(rtc_counter);

    /* RTC counter LSBs.
     * This value is then used by the ANT stack to calculate offset. */
    message_payload[6] = FIRST_BYTE(rtc_counter);            
    message_payload[7] = SECOND_BYTE(rtc_counter);       

    // Broadcast the data as a time sync packet.
    err_code = sd_ant_time_sync_broadcast_tx(ANT_CHANNEL_NUMBER,
                                             ANT_STANDARD_DATA_PAYLOAD_SIZE,
                                             message_payload);
    APP_ERROR_CHECK(err_code);
}

/**
 * @brief Handle ANT events
 */
static void ant_event_handler(ant_evt_t * p_ant_evt, void * p_context)
{
    switch (p_ant_evt->event)
    {
        case EVENT_TX:                      // Fall through
        case EVENT_TRANSFER_TX_COMPLETED:   // Fall through
        case EVENT_TRANSFER_TX_FAILED:
            msg_counter++;
            if (msg_counter == MSGS_BETWEEN_TIMESYNCS)
            {
                // Reset counter
                msg_counter = 0;

                // Send time sync
                send_timesync_message();
            }
            else 
            {
                // Send ACK message
                uint8_t buffer[8] = {0};
                uint32_t err_code = sd_ant_acknowledge_message_tx(ANT_CHANNEL_NUMBER,
                                                                  ANT_STANDARD_DATA_PAYLOAD_SIZE,
                                                                  buffer);
                APP_ERROR_CHECK(err_code);
            }
        break;
        case EVENT_RX: // Do nothing with the messages received
        break;
        default:
        break;
    }
}

// Register the ant_evt_handler as an ANT observer.
NRF_SDH_ANT_OBSERVER(m_ant_observer, APP_ANT_OBSERVER_PRIO, ant_event_handler, NULL);

/**@brief Function for ANT stack initialization.
 *
 * @details Initializes the SoftDevice and the ANT event interrupt.
 */
void softdevice_setup(void)
{
    ret_code_t err_code = nrf_sdh_enable_request();
    APP_ERROR_CHECK(err_code);

    ASSERT(nrf_sdh_is_enabled());

    err_code = nrf_sdh_ant_enable();
    APP_ERROR_CHECK(err_code);
}

int main(void)
{
    // Setup SoftDevice and events handler      
    softdevice_setup();

    // RTC1 init.
    rtc_init();

    // Setup ANT and start channel
    antmaster_ant_setup();

    // Enter main loop
    for (;;)
    {
        // Do nothing
    }
}